"""UDP/IP非同期通信モジュール
"""
import asyncio
from dataclasses import dataclass, field


@dataclass
class Messages:
    send: any = field(init=True)
    receive: any = field(init=False)


class UDPClientConnection:
    def __init__(self, message: Messages, on_con_lost):
        self.message: Messages = message
        self.on_con_lost = on_con_lost
        self.transport = None

    def connection_made(self, transport):
        self.transport = transport
        self.transport.sendto(self.message.send)

    def datagram_received(self, data, addr):
        self.message.receive = data
        self.transport.close()

    def error_received(self, exc):
        print("Error received:", exc)

    def connection_lost(self, exc):
        # self.transport.close()
        self.on_con_lost.set_result(True)


@dataclass
class EtherCATMasterConnection:
    host: str = field(default_factory=str, init=True)
    port: int = field(default=9001, init=True)
    received_data: any = field(default=None, init=False)

    async def send_data(self, message):
        loop = asyncio.get_running_loop()
        on_con_lost = loop.create_future()
        messages = Messages(message)
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: UDPClientConnection(messages, on_con_lost), remote_addr=(self.host, self.port)
        )
        try:
            await asyncio.wait_for(on_con_lost, 3)
            self.received_data = messages.receive
        finally:
            transport.close()



"""
0x9*** 台のMaster ODを定義したモジュール
"""
from dataclasses import field
from pyetg1510.mailbox.sdo_application_interface import SdoEntry, SdoMetadata, SdoDataBody
from pyetg1510.mailbox import SDORequest


class InformationData(SdoDataBody):
    NumberOfEntries: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B"))
    StationAddress: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H"))
    VendorId: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="I"))
    ProductCode: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="I"))
    RevisionNumber: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="I"))
    SerialNumber: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="I"))
    DLStatusRegister: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="I"))


InformationDataFormat = SdoMetadata(
    index=0x9000,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=16,
    request_container=SDORequest,
    response_container=InformationData,
)



"""データ永続化モジュール"""
import os
import json
from collections.abc import Mapping
import copy
from pyetg1510.helper import SysLog

logger = SysLog.logger


class settings:
    def __init__(self, initial_data, *file_name):
        self._fp = None
        self._setting_data = initial_data
        self._origin_data = initial_data
        self._logger = SysLog.logger
        if len(file_name) > 0:
            self._setting_dir = os.path.dirname(file_name[0])
            self._setting_file = os.path.basename(file_name[0])
            self._fullpath = self._setting_dir + "/" + self._setting_file
            self._inititalize()

    @property
    def file(self):
        return self._setting_file

    @file.setter
    def file(self, f):
        self._setting_file = f
        self._fullpath = self._setting_dir + "/" + self._setting_file

    @property
    def dir(self):
        return self._setting_dir

    @dir.setter
    def dir(self, f):
        self._setting_dir = f
        self._fullpath = self._setting_dir + "/" + self._setting_file

    @property
    def data(self):
        return self._setting_data

    def _inititalize(self):
        if not os.path.exists(self._setting_dir):
            os.mkdir(self._setting_dir)
        if os.path.isfile(self._fullpath):
            self.load()
        else:
            self._fp = open(self._fullpath, "w", encoding="utf-8")
            self.save()

    def _deepupdate(self, dict_base, other):
        for k, v in other.items():
            if isinstance(v, Mapping) and k in dict_base:
                self._deepupdate(dict_base[k], v)
            else:
                dict_base[k] = v

    def _report_diff(self, before: dict, after: dict, path: str = ""):
        for k, v in after.items():
            if isinstance(v, Mapping) and k in before:
                if path == "":
                    parent = path
                else:
                    parent = path + "."
                self._report_diff(before[k], v, parent + str(k))
            else:
                if k in before and before[k] != v:
                    self._logger.warning(
                        "Setting changed : " + path + "." + str(k) + " : " + str(before[k]) + " -> " + str(v)
                    )

    def save(self):
        try:
            self._fp = open(self._fullpath, "w", encoding="utf-8")
            self._report_diff(self._origin_data, self._setting_data, "")
            json.dump(
                self._setting_data, self._fp, ensure_ascii=False, indent=4, sort_keys=True, separators=(",", ": ")
            )
            self._origin_data = copy.deepcopy(self._setting_data)
            self._fp.close()
        except Exception as e:
            logger.exception(e)

    def load(self):
        try:
            self._fp = open(self._fullpath, "r", encoding="utf-8")
            self._deepupdate(self._setting_data, json.load(self._fp))
            self._origin_data = copy.deepcopy(self._setting_data)
            self._fp.close()
        except json.JSONDecodeError as e:
            logger.exception(e)
            raise e
        except ValueError as e:
            logger.exception(e)
            raise e
        self.save()

    def make_slice_object(self, start, count):
        s = start - 1
        c = s + count
        return slice(s, c)

    def __del__(self):
        if not self._fp is None:
            self._fp.close()



"""
0x8*** 台のMaster ODを定義したモジュール
"""
from pyetg1510.mailbox.sdo_application_interface import SdoEntry, SdoDataBody, SdoMetadata
from dataclasses import dataclass, field
from pyetg1510.mailbox import SDORequest


@dataclass
class LinkStatus:
    no_link: bool
    """no link"""
    link_no_commm: bool
    """Sub device report a physical link without communication for one or more ports"""
    Link_miss: bool
    """Sub device report a missing link on one or more ports where a link is expected"""
    link_add: bool
    """Sub device report a link on one or more ports where no link is expected"""
    port0: bool
    """one of previous bits refers to port 0"""
    port1: bool
    """one of previous bits refers to port 1"""
    port2: bool
    """one of previous bits refers to port 2"""
    port3: bool
    """one of previous bits refers to port 3"""


@dataclass
class LinkPreset:
    port1_expects_connection: bool
    """Port 1 expects connection to an EtherCAT slave"""
    port2_expects_connection: bool
    """Port 2 expects connection to an EtherCAT slave"""
    port3_expects_connection: bool
    """Port 3 expects connection to an EtherCAT slave"""
    port1_expects_physical_link: bool
    """Port 1 expects physical link"""
    port2_expects_physical_link: bool
    """Port 2 expects physical link"""
    port3_expects_physical_link: bool
    """Port 3 expects physical link"""


class ConfigurationData(SdoDataBody):
    """構成データ オブジェクト 0x8nnn は、サブデバイスのコンフィギュレーションデータが格納されます。
    マスターによって一度だけ更新されます。ホットコネクトグループに属するサブデバイスは最後にリストされます。
    """

    NumberOfEntries: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B"))
    """
    名称:
        エントリ数
    
    """
    FixedStationAddress: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=1, value=0, format="H"))
    """
    名称:
       Fixed Station Address    
    Object:
        0x8nnn:01
    説明:
        `ENI 要素 <https://www.ethercat.org/jp/downloads/downloads_D857E986BC634321840CA5DB43A04D55.htm>`_ の 
        ``Slave:Info:PhysAddr`` の値(各サブデバイスの ESI レジスタのアドレス 0x0010-0x0011 に対応)
    """

    Type: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=2, value="", format="s", size=16))
    """
    名称:
        Type
    Object:
        0x8nnn:02
    説明:
        `ESI 要素 <https://www.ethercat.org/jp/downloads/downloads_6A46D45EA33C47ECB2BB2686BBA963EC.htm>`_ の 
        ``Device:DeviceType:Type`` の値( category General の ``OrderIdx`` [#fn1]_ に対応)
        
    .. [#fn1] `ETG.2010 Slave Information Interface <https://www.ethercat.org/jp/downloads/downloads_AC7ED92675B04BAB9ECD35667A80DFFE.htm>`_  のTable.7に記載
    """
    Name: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=3, value="", format="s", size=32))
    """
    名称:
        Name
    Object:
        0x8nnn:03
    説明:
        `ENI 要素 <https://www.ethercat.org/jp/downloads/downloads_D857E986BC634321840CA5DB43A04D55.htm>`_ の  
        ``Slave:Info:Name`` の値(Object 0x1008 に対応).
    """
    DeviceType: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=4, value=0, format="I"))
    """
    名称:
        Device Type
    Object:
        0x8nnn:04
    説明:
        `ENI 要素 <https://www.ethercat.org/jp/downloads/downloads_D857E986BC634321840CA5DB43A04D55.htm>`_ の 
        ``Device:Profile:ProfileNo (low word)`` と ``Device:Profile:AddInfo (high word)`` の値（object 0x1000に対応）。
        同じ要素が ENI 要素の ``Slave:Mailbox:CoE:ChannelInfo:ProfileNo (low word)`` と 
        ``Slave:Mailbox:CoE:ChannelInfo:AddInfo (high word)`` にも展開されている。
    """
    VendorId: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=5, value=0, format="I"))
    """
    名称:
        Vendor ID, Product Code, Revision, Serial Number
    Object:
        0x8nnn:05-08
    説明:
        `ENI 要素 <https://www.ethercat.org/jp/downloads/downloads_D857E986BC634321840CA5DB43A04D55.htm>`_ の
        ``Slave:Info:VendorId``, ``Slave:Info:ProductCode``, ``Slave:Info:RevisionNo``,
        ``Slave:Info:SerialNo`` の値。 (オブジェクト 0x1018:01-04 に対応).
    """
    ProductCode: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=6, value=0, format="I"))
    RevisionNumber: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=7, value=0, format="I"))
    SerialNumber: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=8, value=0, format="I"))
    MailboxOutSize: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=33, value=0, format="H"))
    """Values of ENI elements Slave:Mailbox:Send:Length and Slave:Mailbox:Receive:Length of slave n
    (correspond to ESC registers 0x0802-0x0803 and 0x080A-0x080B)."""
    MailboxInSize: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=34, value=0, format="H"))
    LinkStatus: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=35, value=0, format="B"))
    """Link status of sub terminal which are also able to get link_preset property"""
    LinkPreset: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=36, value=0, format="B"))
    """Reports the expected physical link on ports 1, 2 and 3 of sub terminal which are also able to get link_preset property"""
    Flags: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=37, value=0, format="B"))
    """Provides additional topology information about redundancy and hot connect"""
    PortPhysics: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=38, value=0, format="H"))
    """Corresponds to ESI element Device:Physics and to ENI element sub:Info:Physics."""
    MailboxProtocolsSupported: SdoEntry = field(
        default_factory=lambda: SdoEntry[int](sub_index=39, value=0, format="H")
    )
    DiagHistoryObjectSupported: SdoEntry = field(
        default_factory=lambda: SdoEntry[bool](sub_index=40, value=False, format="?")
    )

    @property
    def link_status(self) -> LinkStatus:
        args = [bool(self.LinkStatus.value & 1 << i) for i in range(0, 8)]
        return LinkStatus(*args)

    @property
    def link_preset(self) -> LinkPreset:
        args = [bool(self.LinkPreset.value & 1 << i) for i in range(0, 3)]
        args.extend([bool(self.LinkPreset.value & 16 << i) for i in range(0, 3)])
        return LinkPreset(*args)

    @property
    def redundancy_adapter_port(self) -> int:
        """Specified port is connected to secondary master adapter for redundancy purposes.
        zero is not used.
        """
        return self.Flags.value & 0x3

    @property
    def hot_connect_head_terminal(self) -> bool:
        """This terminal is head terminal of hot connect group"""
        return bool(self.Flags.value & 0x4)

    @property
    def hot_connect(self) -> bool:
        """This terminal belong to hot connect group"""
        return bool(self.Flags.value & 0x8)

    @property
    def aoe_supported(self) -> bool:
        return bool(self.MailboxProtocolsSupported.value & 1)

    @property
    def eoe_supported(self) -> bool:
        return bool(self.MailboxProtocolsSupported.value & 2)

    @property
    def coe_supported(self) -> bool:
        return bool(self.MailboxProtocolsSupported.value & 4)

    @property
    def foe_supported(self) -> bool:
        return bool(self.MailboxProtocolsSupported.value & 8)

    @property
    def soe_supported(self) -> bool:
        return bool(self.MailboxProtocolsSupported.value & 16)

    @property
    def voe_supported(self) -> bool:
        return bool(self.MailboxProtocolsSupported.value & 32)


ConfigurationDataFormat = SdoMetadata(
    index=0x8000,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=16,
    request_container=SDORequest,
    response_container=ConfigurationData,
)



"""
0xf*** 台のMaster ODを定義したモジュール
"""
from pyetg1510.mailbox.sdo_application_interface import SdoEntry, SdoMetadata, SdoDataBody
from dataclasses import field
from pyetg1510.mailbox import SDORequest


class DetectModulesCommand(SdoDataBody):
    NumberofEntries: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B"))
    ScanCommandRequest: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=1, value="", format="2s"))
    ScanCommandStatus: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=2, value=0, format="B"))
    ScanCommandResponse: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=3, value="", format="6s"))


class ConfiguredAddressList(SdoDataBody):
    NumberofSlaves: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B"))
    ConfiguredAddress: SdoEntry = field(default_factory=lambda: SdoEntry[list](sub_index=1, value=[0] * 125, format="H"))


class MasterDiagData(SdoDataBody):
    NumberOfEntries: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B"))
    CyclicLostFrames: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=1, value=0, format="I"))
    ACyclicLostFrames: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=2, value=0, format="I"))
    CyclicFramesPerSecond: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=3, value=0, format="I"))
    ACyclicFramesPerSecond: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=4, value=0, format="I"))
    MasterState: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=16, value=0, format="H"))


class DiagInterfaceControl(SdoDataBody):
    NumberOfEntries: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B"))
    ResetDiagInfo: SdoEntry = field(default_factory=lambda: SdoEntry[bool](sub_index=16, value=0, format="?"))


DetectModulesCommandFormat = SdoMetadata(
    index=0xF002,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=3,
    request_container=SDORequest,
    response_container=DetectModulesCommand,
)

ConfiguredAddressListFormat = SdoMetadata(
    index=0xF020,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=0x7F,
    request_container=SDORequest,
    response_container=ConfiguredAddressList,
)


MasterDiagDataFormat = SdoMetadata(
    index=0xF120,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=16,
    request_container=SDORequest,
    response_container=MasterDiagData,
)

DiagInterfaceControlFormat = SdoMetadata(
    index=0xF200,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=16,
    request_container=SDORequest,
    response_container=DiagInterfaceControl,
)



"""

ETG.1510データ収集制御モジュール。 ``sdo_*xxx_`` から始まるモジュールで定義されたメタデータを通じてデータモデルを参照。
"""
from dataclasses import dataclass, field, fields
from pyetg1510.mailbox.connection import EtherCATMasterConnection
from pyetg1510.mailbox.sdo_application_interface import (
    SdoDataBody,
    ConcreteSDODataFactory,
    is_primitive,
    SdoMetadataMapper,
    MappingMember,
    SdoDataController,
    ODListFormat,
    SDOInfoDescriptionFormat,
    SDOInfoEntryFormat,
)
from pyetg1510.sdo_1xxx_master_object import (
    DeviceTypeFormat,
    DeviceNameFormat,
    HardwareVersionFormat,
    SoftwareVersionFormat,
    IndentityObjectFormat,
)
from pyetg1510.sdo_8xxx_configuration_data import ConfigurationDataFormat
from pyetg1510.sdo_9xxx_information_data import InformationDataFormat
from pyetg1510.sdo_axxx_master_diagnosis import DiagnosisDataFormat
from pyetg1510.sdo_fxxx_controls import (
    DetectModulesCommandFormat,
    MasterDiagDataFormat,
    DiagInterfaceControlFormat,
    ConfiguredAddressListFormat,
)
from typing import List, Tuple
from pyetg1510.helper import SysLog

logger = SysLog.logger


class MasterDiagnosisMetadataMapper(SdoMetadataMapper):
    """定義された :obj:`メタデータ <pyetg1510.mailbox.sdo_data_factory.SdoMetadata>` とそのインデックス範囲を関連付けるクラス

    ベースクラス :mod:`pyetg1510.mailbox.sdo_data_factory.SdoMetadataMapper.find` メソッドにより何れかのメンバーをセレクトすることができる。
    """

    device_type = MappingMember(index_range=(0x1000, 0x1000), metadata=DeviceTypeFormat)
    device_name = MappingMember(index_range=(0x1008, 0x1008), metadata=DeviceNameFormat)
    hardware_vaersion = MappingMember(index_range=(0x1009, 0x1009), metadata=HardwareVersionFormat)
    software_vaersion = MappingMember(index_range=(0x100A, 0x100A), metadata=SoftwareVersionFormat)
    indentity_object = MappingMember(index_range=(0x1018, 0x1018), metadata=IndentityObjectFormat)
    master_configuration = MappingMember(index_range=(0x8000, 0x8FFF), metadata=ConfigurationDataFormat)
    master_information = MappingMember(index_range=(0x9000, 0x9FFF), metadata=InformationDataFormat)
    diagnosis_data = MappingMember(index_range=(0xA000, 0xAFFF), metadata=DiagnosisDataFormat)
    detect_modules_command = MappingMember(index_range=(0xF002, 0xF002), metadata=DetectModulesCommandFormat)
    configured_address_list = MappingMember(index_range=(0xF020, 0xF020), metadata=ConfiguredAddressListFormat)
    master_diag_data = MappingMember(index_range=(0xF120, 0xF120), metadata=MasterDiagDataFormat)
    diag_interface_control = MappingMember(index_range=(0xF200, 0xF200), metadata=DiagInterfaceControlFormat)


@dataclass
class MasterODSpecification:
    """SDO Information service [#f1]_  により、収集可能な、EtherCAT main device のサポートするSDO情報を管理するクラス。

    .. [#f1] `ETG.1000.6 <https://www.ethercat.org/jp/downloads/downloads_A02E436C7A97479F9261FDFA8A6D71E5.htm>`_ Section 5.6.3

    Args:
        connection(EtherCATMasterConnection): 通信コネクタオブジェクト

    """

    connection: EtherCATMasterConnection
    sdo_data_entity: ConcreteSDODataFactory = field(default_factory=ConcreteSDODataFactory, init=False)

    def __post_init__(self):
        self.data_handler = SdoDataController(session=self.connection, get_info=True)
        self.current_index = 0
        self.current_subindex = 0

    async def get_object_dictionary(self):
        """SDO Information serviceによりmain deviceのODを問い合わせ、その仕様をsdo_data_entryへ登録。

        次の手順で実行する。

        1. OD Listを収集し、 :obj:`メタデータマッパ <pyetg1510.etg_1510.MasterDiagnosisMetadataMapper>` を用いて
           対応する :obj:`メタデータ <pyetg1510.mailbox.sdo_data_factory.SdoMetadata>` を取得
        2. :mod:`ConcreteSDODataFactory` によりテンプレートを元に実体を作成し、sdo_data_entryに登録。
        3. 個々のODのDescriptionを要求し、作成した実体をレスポンスに従い更新。
        4. sdo_data_entryに問い合わせたデータが作成される。

        """
        logger.info("Fetch OD List")
        await self.data_handler.fetch(sdo_metadata=ODListFormat, sdo_data=ODListFormat.response_container())

        for self.current_index in self.data_handler.sdo_data.ObjectIndex.value:
            logger.info(f"==== Index {self.current_index}, format :{ODListFormat.response_container.__name__}")
            _selected: MappingMember = MasterDiagnosisMetadataMapper.find(self.current_index)
            if _selected is not None:
                # Create instances by sdo factory.
                self.sdo_data_entity.create(index=self.current_index, template=_selected.metadata.response_container)
            else:
                logger.warning(f"Index : {self.current_index} is not defined for any specification.")
                continue
            SDOInfoDescriptionFormat.index = self.current_index
            logger.info("Fetch Object Description")
            await self.data_handler.fetch(
                sdo_metadata=SDOInfoDescriptionFormat, sdo_data=SDOInfoDescriptionFormat.response_container()
            )
            # apply max subindex to configuration model
            if hasattr(self.data_handler.sdo_data, "MaxSubindex"):
                _selected.metadata.max_sub_index = self.data_handler.sdo_data.MaxSubindex.value
            else:
                _selected.metadata.max_sub_index = 0
            logger.info(f"Max sub index: {_selected.metadata.max_sub_index}")
            for field in fields(self.sdo_data_entity.entries[self.current_index]):
                entry = getattr(self.sdo_data_entity.entries[self.current_index], field.name)
                self.current_subindex = entry.sub_index
                logger.info(f"   ---- Subindex {self.current_subindex}")
                SDOInfoEntryFormat.index = self.current_index
                SDOInfoEntryFormat.sub_index = self.current_subindex
                logger.info("Fetch Entry Description")
                await self.data_handler.fetch(
                    sdo_metadata=SDOInfoEntryFormat, sdo_data=SDOInfoEntryFormat.response_container()
                )
                if "AbortCode" not in self.data_handler.sdo_data.__dict__:
                    if entry is not None:
                        entry.name = self.data_handler.sdo_data.Data.value
                        if is_primitive(entry.value):
                            entry.size = int(self.data_handler.sdo_data.BitLength.value / 8)
                        entry.enable = True
                        logger.info(
                            f"Index: {self.current_index}. Subindex:{self.current_subindex} is Enabled. Size:{entry.size}"
                        )
                        logger.info(f"     {entry}")
                    else:
                        logger.error(f"{self.current_index}:{self.current_subindex} is not found on definition.")

        logger.info("============ Information data fetch complete ==============")


@dataclass
class ETG1510Profile:
    """ETG.1510 データ収集非同期イテレータ

    master_od にセットしたODリストに従い、順次SDOを収集して返すイテレータクラス。

    使用例:
        .. code-block:: python

            async def get_sdo_data():
                # create iterator object
                ittr = ETG1510Profile(master_od=<<MasterConfigurationクラスで作成したODデータモデル>>)
                async for sdo in ittr:
                    # sdo will be possible to get each SDO as SdoDataBody type or Dictionary type.
                    pass

    Args:
        master_od(MasterODSpecification): :meth:`get_object_dictionaryメソッド <pyetg1510.etg_1510.MasterODSpecification.get_object_dictionary>`
                                        を実行してODを収集完了した後のMasterODSpecificationオブジェクト
        watch_index_list(List[int]): 監視対象のSDOインデックスリスト。未定義の場合はOD全て対象。

    Return:
        Tuple[int, SdoDataBody]: SDOインデックス, 取得したSDOデータコンテナ
    """

    master_od: MasterODSpecification
    """収集したメインデバイスのオブジェクトディクショナリ"""
    watch_index_list: List[int] = None
    """イテレータで収集する際に、収集対象となるインデックスリストを指定する場合はそのリストを設定する。指定しない場合は全て返す"""

    def __post_init__(self):
        self.data_handler = SdoDataController(session=self.master_od.connection, get_info=False)
        self.watch_address = 0
        self.sdo_database = self.master_od.sdo_data_entity.entries

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.watch_index_list is None:
            sdo_index_list = list(self.master_od.sdo_data_entity.entries.keys())
        else:
            sdo_index_list = self.watch_index_list
        if len(sdo_index_list) <= self.watch_address:
            self.watch_address = 0
            raise StopAsyncIteration
        # ToDo: sdo_index_listの要素に self.master_od.sdo_data_entity.entries.keys() が含まれなければ異常終了する
        sdo_metadata = MasterDiagnosisMetadataMapper.find(sdo_index_list[self.watch_address]).metadata
        sdo_metadata.index = sdo_index_list[self.watch_address]
        logger.info(f"==== Fetch and update data index:{sdo_index_list[self.watch_address]}")
        await self.data_handler.fetch(
            sdo_metadata=sdo_metadata, sdo_data=self.sdo_database[sdo_index_list[self.watch_address]]
        )
        report = (sdo_index_list[self.watch_address], self.data_handler.sdo_data)
        self.watch_address += 1
        return report

    async def get_sdo(self, index: int) -> SdoDataBody:
        """
        指定したインデックスのSDOを取得する

        Args:
            index(int): SDOインデックス

        Return:
            SdoDataBody: 取得したSDOデータコンテナ
        """

        sdo_metadata = MasterDiagnosisMetadataMapper.find(index).metadata
        sdo_metadata.index = index
        await self.data_handler.fetch(sdo_metadata=sdo_metadata, sdo_data=self.sdo_database[index])

        return self.data_handler.sdo_data



# This is a sample Python script.
from __future__ import annotations
from pyetg1510.mailbox.sdo_application_interface import SdoMetadata
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import List
from pyetg1510.helper import SysLog

logger = SysLog.logger


@dataclass
class Component(metaclass=ABCMeta):
    """
    The base Component class declares common operations for both simple and
    complex objects of a composition.
    """

    parent: Component = field(default=None)

    def __new__(cls, *args, **kwargs):
        dataclass(cls)
        return super().__new__(cls)

    @abstractmethod
    def add(self, component: Component) -> None:
        pass

    @abstractmethod
    def remove(self, component: Component) -> None:
        pass

    @abstractmethod
    def operation(self) -> str:
        """
        The base Component may implement some default behavior or leave it to
        concrete classes (by declaring the method containing the behavior as
        "abstract").
        """
        pass

    @abstractmethod
    def is_composite(self) -> bool:
        pass


class Leaf(Component, SdoMetadata):
    """
    The Leaf class represents the end objects of a composition. A leaf can't
    have any children.

    Usually, it's the Leaf objects that do the actual work, whereas Composite
    objects only delegate to their sub-components.
    """

    @property
    def upstream(self):
        if self.parent is not None:
            current_index = self.parent.children.index(self)
            if current_index > 0:
                return self.parent.children[current_index - 1]
            else:
                return None
        else:
            return None

    @property
    def downstream(self):
        if self.parent is not None:
            current_index = self.parent.children.index(self)
            if current_index < len(self.parent.children):
                return self.parent.children[current_index + 1]
            else:
                return None
        else:
            return None

    def is_composite(self) -> bool:
        """
        You can provide a method that lets the client code figure out whether a
        component can bear children.
        """

        return False


class Composite(Component):
    """
    The Composite class represents the complex components that may have
    children. Usually, the Composite objects delegate the actual work to their
    children and then "sum-up" the result.
    """

    children: List[Component] = field(default_factory=list, init=True)

    """
    A composite object can add or remove other components (both simple or
    complex) to or from its child list.
    """

    def add(self, component: Component) -> None:
        self.children.append(component)
        component.parent = self

    def remove(self, component: Component) -> None:
        self.children.remove(component)
        component.parent = None

    def is_composite(self) -> bool:
        return True

    def operation(self) -> str:
        """
        The Composite executes its primary logic in a particular way. It
        traverses recursively through all its children, collecting and summing
        their results. Since the composite's children pass these calls to their
        children and so forth, the whole object tree is traversed as a result.
        """

        results = []
        for child in self.children:
            results.append(child.operation())
        return f"Branch({'+'.join(results)})"



"""
0xa*** 台のMaster ODを定義したモジュール
"""
from dataclasses import field, dataclass
from pyetg1510.mailbox.sdo_application_interface import SdoEntry, SdoMetadata, SdoDataBody
from typing import List
from enum import Enum
from pyetg1510.mailbox import SDORequest


@dataclass
class ALStatusCodeDef:
    """Alarm definition member"""
    code: int
    """Alarm code"""
    occurrence_timing: str
    """Watch timing"""
    transition_state: str
    """Transition of state machine"""
    reference: str
    """Documentation for solution"""

class ALStausCode(Enum):
    """Definition of alarm status code"""
    NoError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0000, occurrence_timing="Any", transition_state="Current", reference="ETG.1000.6"
    )
    UnspecifiedError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0001, occurrence_timing="Any", transition_state="Any +E", reference="ETG.1000.6"
    )
    NoMemory: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0002, occurrence_timing="Any", transition_state="Any +E", reference="ETG.1000.6"
    )
    InvalidRevision: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0004, occurrence_timing="PS", transition_state="P +E", reference="Additionalcode"
    )
    InvalidDeviceSetup: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0003, occurrence_timing="PS", transition_state="P +E", reference="Additionalcode"
    )
    Sii_EEPROMInformationDoesNotMatchFirmware: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0006, occurrence_timing="IP", transition_state="I +E", reference="Additionalcode"
    )
    FirmwareUpdateNotSuccessful_OldFirmwareStillRunning: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0007, occurrence_timing="Boot", transition_state="I +E", reference="Additionalcode"
    )
    LicenseError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x000E, occurrence_timing="Any", transition_state="I +E", reference="Additionalcode"
    )
    InvalidRequestedStateChange: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0011, occurrence_timing="Any", transition_state="Current +E(not O +E)", reference="ETG.1000.6"
    )
    UnknownRequestedState: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0012, occurrence_timing="Any", transition_state="Current +E(not O +E)", reference="ETG.1000.6"
    )
    BootstrapNotSupported: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0013, occurrence_timing="IB", transition_state="I +E", reference="ETG.1000.6"
    )
    NoValidFirmware: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0014, occurrence_timing="IP", transition_state="I +E", reference="ETG.1000.6"
    )
    InvalidMailboxConfigurationBOOT: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0015, occurrence_timing="IB", transition_state="I +E", reference="ETG.1000.6"
    )
    InvalidMailboxConfigurationPREOP: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0016, occurrence_timing="IP", transition_state="I +E", reference="ETG.1000.6"
    )
    InvalidSyncManagerConfiguration: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0017, occurrence_timing="PS, SO", transition_state="Current +E(not O +E)", reference="ETG.1000.6"
    )
    NoValidInputsAvailable: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0018, occurrence_timing="O, SO", transition_state="S +E", reference="ETG.1000.6"
    )
    NoValidOutputs: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0019, occurrence_timing="O, SO", transition_state="S +E", reference="ETG.1000.6"
    )
    SynchronizationError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x001A, occurrence_timing="O, SO", transition_state="S +E", reference="ETG.1000.6"
    )
    SyncManagerWatchdog: ALStatusCodeDef = ALStatusCodeDef(
        code=0x001B, occurrence_timing="O, S", transition_state="S +E", reference="ETG.1000.6"
    )
    InvalidSyncManagerTypes: ALStatusCodeDef = ALStatusCodeDef(
        code=0x001C, occurrence_timing="O, S, PS", transition_state="S +E", reference="ETG.1000.6"
    )
    InvalidOutputConfiguration: ALStatusCodeDef = ALStatusCodeDef(
        code=0x001D, occurrence_timing="O, S, PS", transition_state="S +E", reference="ETG.1000.6"
    )
    InvalidInputConfiguration: ALStatusCodeDef = ALStatusCodeDef(
        code=0x001E, occurrence_timing="O, S, PS", transition_state="S +E", reference="ETG.1000.6"
    )
    InvalidWatchDogConfiguration: ALStatusCodeDef = ALStatusCodeDef(
        code=0x001F, occurrence_timing="O, S, PS", transition_state="P +E", reference="ETG.1000.6"
    )
    SlaveNeedsColdstart: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0020, occurrence_timing="Any", transition_state="Current +E(not O +E)", reference="ETG.1000.6"
    )
    SlaveNeedsInit: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0021, occurrence_timing="B, P, S, O", transition_state="Current +E(not O +E)", reference="ETG.1000.6"
    )
    SlaveNeedsPREOP: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0022, occurrence_timing="S, O", transition_state="Current +E(not O +E)", reference="ETG.1000.6"
    )
    SlaveNeedsSAFEOP: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0023, occurrence_timing="O", transition_state="Current +E(not O +E)", reference="ETG.1000.6"
    )
    InvalidInputMapping: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0024, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    InvalidOutputMapping: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0025, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    InconsistentSettings: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0026, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    FreeRunNotSupported: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0027, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    SynchronizationNotSupported: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0028, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    FreeRunNeeds3BufferMode: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0029, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    BackgroundWatchdog: ALStatusCodeDef = ALStatusCodeDef(
        code=0x002A, occurrence_timing="S, O", transition_state="P +E", reference="ETG.1000.6"
    )
    NoValidInputsAndOutputs: ALStatusCodeDef = ALStatusCodeDef(
        code=0x002B, occurrence_timing="O, SO", transition_state="S +E", reference="ETG.1000.6"
    )
    FatalSyncError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x002C, occurrence_timing="O", transition_state="S +E", reference="ETG.1000.6"
    )
    NoSyncError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x002D, occurrence_timing="SO", transition_state="S +E", reference="ETG.1000.6"
    )
    CycleTimeTooSmall: ALStatusCodeDef = ALStatusCodeDef(
        code=0x002E, occurrence_timing="SO", transition_state="S +E", reference="AdditionalCode"
    )
    InvalidDcSyncConfiguration: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0030, occurrence_timing="O, SO, PS", transition_state="P +E,S +E", reference="ETG.1000.6"
    )
    InvalidDcLatchConfiguration: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0031, occurrence_timing="O, SO, PS", transition_state="P +E,S +E", reference="ETG.1000.6"
    )
    PllError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0032, occurrence_timing="S, O", transition_state="S +E", reference="ETG.1000.6"
    )
    DcSyncIoError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0033, occurrence_timing="O, SO", transition_state="S +E", reference="ETG.1000.6"
    )
    DcSyncTimeoutError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0034, occurrence_timing="O, SO", transition_state="S +E", reference="ETG.1000.6"
    )
    DcInvalidSyncCycleTime: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0035, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    DcSync0CycleTime: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0036, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    DcSync1CycleTime: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0037, occurrence_timing="PS", transition_state="P +E", reference="ETG.1000.6"
    )
    Mbx_Aoe: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0041, occurrence_timing="B, P, S,O", transition_state="Current +ES +E", reference="ETG.1000.6"
    )
    Mbx_Eoe: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0042, occurrence_timing="B, P, S,O", transition_state="Current +ES +E", reference="ETG.1000.6"
    )
    Mbx_Coe: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0043, occurrence_timing="B, P, S,O", transition_state="Current +ES +E", reference="ETG.1000.6"
    )
    Mbx_Foe: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0044, occurrence_timing="B, P, S,O", transition_state="Current +ES +E", reference="ETG.1000.6"
    )
    Mbx_Soe: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0045, occurrence_timing="B, P, S,O", transition_state="Current +ES +E", reference="ETG.1000.6"
    )
    Mbx_Voe: ALStatusCodeDef = ALStatusCodeDef(
        code=0x004F, occurrence_timing="B, P, S,O", transition_state="Current +ES +E", reference="ETG.1000.6"
    )
    EepromNoAccess: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0050, occurrence_timing="Any", transition_state="Any +E(not O +E)", reference="ETG.1000.6"
    )
    EepromError: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0051, occurrence_timing="Any", transition_state="Any +E(not O +E)", reference="ETG.1000.6"
    )
    ExternalHardwareNotReady: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0052, occurrence_timing="Any", transition_state="Any +E(not O +E)", reference="Additionalcode"
    )
    SlaveRestartedLocally: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0060, occurrence_timing="Any", transition_state="I", reference="ETG.1000.6"
    )
    DeviceIdentificationValueUpdated: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0061, occurrence_timing="P", transition_state="P +E", reference="ETG.1000.6"
    )
    DetectedModuleIdentListDoesNotMatch: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0070, occurrence_timing="PS", transition_state="P +E", reference="Additionalcode"
    )
    SupplyVoltageTooLow: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0080, occurrence_timing="Any", transition_state="Any +E(not O +E)", reference="Additionalcode"
    )
    SupplyVoltageTooHigh: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0081, occurrence_timing="Any", transition_state="Any +E(not O +E)", reference="Additionalcode"
    )
    TemperatureTooLow: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0082, occurrence_timing="Any", transition_state="Any +E(not O +E)", reference="Additionalcode"
    )
    TemperatrueTooHigh: ALStatusCodeDef = ALStatusCodeDef(
        code=0x0083, occurrence_timing="Any", transition_state="Any +E(not O +E)", reference="Additionalcode"
    )
    ApplicationControllerAvailable: ALStatusCodeDef = ALStatusCodeDef(
        code=0x00F0, occurrence_timing="I", transition_state="I +E", reference="Additionalcode"
    )

    @classmethod
    def get_al(cls, code: int):
        for obj in cls:
            if code == obj.value.code:
                return obj
        return None


class ALStatus(Enum):
    INIT = 0x0001
    PREOP = 0x0002
    SAFEOP = 0x0004
    OP = 0x0008
    REJECTED = 0x0010
    ALCODE_UPDATED = 0x0020


class LoopControl(Enum):
    """Value of loop control of port X"""
    Auto: int = 0
    """Port loop opens automatically when a physical link is established, and closes automatically when the physical link is lost"""
    AutoClose: int = 1
    """Port loop closes when the physical link is lost, and opens when the physical link is established after explicit request from master"""
    Open: int = 2
    """Port loop is always open, independently from the physical link state"""
    Close: int = 3
    """Port loop is always closed, independently from the physical link state"""

    @classmethod
    def find(cls, value: int):
        print(f"Find value:{value}")
        for e in LoopControl:
            if e.value == value:
                return e
        raise ValueError(f"{value} is wrong value.")


@dataclass
class PortStatus:
    """Structure of link connection

    use_to_communication=False, link_up=True:
       Used for redundancy

    use_to_communication=True, link_up=True:
       Used for communication with other sub device

    loop_control:
       If not set to Auto, port would not be connected with next port (loop control)
    """
    use_to_communication: bool
    link_up: bool
    loop_control: LoopControl


class DiagnosisData(SdoDataBody):
    """0xAxxx: Diagnosis data of ETG.1510"""
    NumberOfEntries: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B"))
    ALStatus: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=1, value=0, format="H"))
    ALControl: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=2, value=0, format="H"))
    ALStatusCode: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=3, value=0, format="H"))
    LinkConnStatus: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=4, value=0, format="B"))
    LinkControl: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=5, value=0, format="B"))
    FixedAddressConnPort: SdoEntry = field(
        default_factory=lambda: SdoEntry[list](sub_index=6, value=[0, 0, 0, 0], size=8, format="H")
    )
    FrameErrorCounterPort: SdoEntry = field(
        default_factory=lambda: SdoEntry[list](sub_index=10, value=[0, 0, 0, 0], size=16, format="I")
    )
    CyclicWCErrorCounter: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=14, value=0, format="I"))
    SlaveNotPresentCounter: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=15, value=0, format="I"))
    AbnormalStateChangeCounter: SdoEntry = field(
        default_factory=lambda: SdoEntry[int](sub_index=16, value=0, format="I")
    )
    DisableAutomaticLinkControl: SdoEntry = field(
        default_factory=lambda: SdoEntry[bool](sub_index=17, value=False, format="?")
    )
    LastProtocolError: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=18, value=0, format="I"))
    NewDiagMessageAvailable: SdoEntry = field(
        default_factory=lambda: SdoEntry[bool](sub_index=19, value=False, format="?")
    )

    @property
    def port_status(self) -> List[PortStatus]:
        """Ports status of sub device"""
        result = [
            PortStatus(
                bool(self.LinkConnStatus.value & 1 << p),
                bool(self.LinkConnStatus.value & 16 << p),
                LoopControl.find((self.LinkControl.value & 3 << p * 2) >> p * 2),
            )
            for p in range(0, 4)
        ]
        return result

    @property
    def al_status_code(self) -> ALStatus:
        """Alarm status of sub device"""
        return ALStausCode.get_al(self.ALStatusCode.value)

    @property
    def al_control(self) -> ALStatus:
        """AL status: Main device control status of state machine"""
        return ALStatus(0x0F & self.ALControl.value)

    @property
    def al_status(self) -> ALStatus:
        """AL status: Sub device current status of state machine"""
        return ALStatus(0x0F & self.ALStatus.value)

    @property
    def is_rejected(self) -> bool:
        """AL Control status : Update rejected"""
        return bool(self.ALControl.value & ALStatus.REJECTED.value)

    @property
    def is_updated(self) -> bool:
        """AL Control status : Update successfully"""
        return bool(self.ALControl.value & ALStatus.ALCODE_UPDATED.value)


DiagnosisDataFormat = SdoMetadata(
    index=0xA000,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=32,
    request_container=SDORequest,
    response_container=DiagnosisData,
)



from pyetg1510.mailbox.connection import *
from pyetg1510.mailbox.mailbox_gateway import *
from pyetg1510.mailbox.sdo_application_interface import *

VERSION = (0, 0, 1)

__version__ = ".".join([str(x) for x in VERSION])



"""
0x1*** 台のMaster ODを定義したモジュール
"""
from pyetg1510.mailbox.sdo_application_interface import SdoEntry, SdoDataBody, SdoMetadata
from dataclasses import field
from pyetg1510.mailbox import SDORequest


class DeviceTypeData(SdoDataBody):
    """Contains the Device Type of EtherCAT Master device."""

    DeviceType: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="I"))


class DeviceNameData(SdoDataBody):
    """Contains the Manufacturer Device Name of EtherCAT Master device."""

    DeviceName: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=0, value="", format="s"))


class HardwareVersionData(SdoDataBody):
    """Contains the Manufacturer Hardware Version of EtherCAT Master device."""

    HardwareVersion: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=0, value="", format="s"))


class SoftwareVersionData(SdoDataBody):
    """Contains the Manufacturer Software Version of EtherCAT Master device."""

    SoftwareVersion: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=0, value="", format="s"))


class IndentityObjectData(SdoDataBody):
    """Contains the identification parameter of EtherCAT Master device."""

    NumberOfEntries: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H"))
    """Number of entries of 0x1xxx entry."""
    VendorID: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=1, value=0, format="I"))
    """EtherCAT Vendor ID of master manufacturer."""
    ProductCode: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=2, value=0, format="I"))
    """Code uniquely identifying the specific EtherCAT Master device."""
    RevisionNumber: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=3, value=0, format="I"))
    """Revision of EtherCAT Master device."""
    SerialNumber: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=4, value=0, format="I"))
    """Optional serial number of EtherCAT Master Device (set to 0 if not used)."""


DeviceTypeFormat = SdoMetadata(
    index=0x1000,
    sub_index=0x0000,
    support_complete_access=False,
    max_sub_index=0,
    request_container=SDORequest,
    response_container=DeviceTypeData,
)

DeviceNameFormat = SdoMetadata(
    index=0x1008,
    sub_index=0x0000,
    support_complete_access=False,
    max_sub_index=0,
    request_container=SDORequest,
    response_container=DeviceNameData,
)

HardwareVersionFormat = SdoMetadata(
    index=0x1009,
    sub_index=0x0000,
    support_complete_access=False,
    max_sub_index=0,
    request_container=SDORequest,
    response_container=HardwareVersionData,
)


SoftwareVersionFormat = SdoMetadata(
    index=0x100A,
    sub_index=0x0000,
    support_complete_access=False,
    max_sub_index=0,
    request_container=SDORequest,
    response_container=SoftwareVersionData,
)

IndentityObjectFormat = SdoMetadata(
    index=0x1018,
    sub_index=0x0000,
    support_complete_access=True,
    max_sub_index=0,
    request_container=SDORequest,
    response_container=IndentityObjectData,
)



"""Python Logging 制御モジュール"""
from logging.handlers import SysLogHandler
from logging.handlers import RotatingFileHandler
from logging import handlers
import logging
import socket
from enum import Enum
import os


class LoggingLevel(Enum):
    ALL: logging = logging.NOTSET
    DEBUG: logging = logging.DEBUG
    INFO: logging = logging.INFO
    WARNING: logging = logging.WARNING
    ERROR: logging = logging.ERROR
    CRITICAL: logging = logging.CRITICAL
    DISABLE: None = None


class SysLog:
    address: str = "127.0.0.1"
    port: int = handlers.SYSLOG_UDP_PORT
    facility: int = handlers.SysLogHandler.LOG_USER
    socktype: socket = socket.SOCK_DGRAM
    file_path: os.PathLike = "../../soft.log"
    app_name: str = __name__
    logger: logging.Logger = logging.getLogger(app_name)

    @classmethod
    def console_log_configuration(cls, level: LoggingLevel = logging.CRITICAL):
        log_level = level.value
        if log_level is not None:
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(
                logging.Formatter("%(asctime)s : %(levelname)s : %(module)s : %(lineno)d : %(message)s")
            )
            stream_handler.setLevel(log_level)
            cls.logger.addHandler(stream_handler)

    @classmethod
    def rotation_log_configuration(cls, level: LoggingLevel = logging.CRITICAL):
        log_level = level.value
        if log_level is not None:
            filename = cls.file_path
            file_handler = RotatingFileHandler(filename=filename, maxBytes=100000, backupCount=10, encoding="utf-8")

            file_handler.setFormatter(
                logging.Formatter("%(asctime)s : %(levelname)s : %(module)s : %(lineno)d : %(message)s")
            )
            file_handler.setLevel(log_level)
            cls.logger.addHandler(file_handler)

    @classmethod
    def syslog_configuration(cls, level: LoggingLevel = logging.CRITICAL):
        log_level = level.value
        if log_level is not None:
            formatter = logging.Formatter(cls.app_name + ": %(levelname)s : %(module)s : %(lineno)d : %(message)s")

            syslog_handler = SysLogHandler(
                address=(cls.address, cls.port),
                facility=cls.facility,
                socktype=cls.socktype,
            )
            syslog_handler.setFormatter(formatter)
            syslog_handler.setLevel(log_level)
            cls.logger.addHandler(syslog_handler)

    @classmethod
    def set_loglevel(cls, level: LoggingLevel = logging.CRITICAL):
        cls.logger.setLevel(level.value)



"""
SDOデータ生成モジュール
"""
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, asdict, field
import dataclasses
from typing import Dict, Generic, TypeVar, Tuple
from struct import calcsize, unpack_from, unpack, error
from ctypes import Structure
from pyetg1510.helper import SysLog
from pyetg1510.mailbox import (
    EtherCATMasterConnection,
    SDOInformationODListRequest,
    SDOInformationEntryRequest,
    SDOInformationDescriptionRequest,
    SDOResponseMessage,
    SdoService,
    SDORequestInfoMessage,
    SdoInfoOpcode,
    SDOCommandMessage,
)

logger = SysLog.logger

T = TypeVar("T")
is_primitive = lambda x: isinstance(x, (int, float, bool, str))


@dataclass
class AbstructSdoDataBody(metaclass=ABCMeta):
    """SDODataBody インターフェース"""

    def __new__(cls, *args, **kwargs):
        dataclass(cls)
        return super().__new__(cls)

    @property
    @abstractmethod
    def unpack_format(self):
        """unpack format abstract method"""
        pass

    @property
    @abstractmethod
    def total_size(self):
        """total size property get"""
        pass


class SDODataFactory(metaclass=ABCMeta):
    """SDO Data Factory 抽象クラス"""

    def create(self, index: int, template: AbstructSdoDataBody.__class__):
        """SDO data 生成メソッド
        SDODataBodyオブジェクトを生成してインデックスに対応する辞書へ登録する抽象クラス

        Args:
            index(int): SDOインデックス
            template(AbstructSdoDataBody.__class__): :obj:`SDOメタデータ <pyetg1510.mailbox.sdo_data_factory.SdoMetadata>` に定義された
              :obj:`response_container <pyetg1510.mailbox.sdo_data_factory.SdoMetadata.response_container>` に定義したテンプレートクラス
        """
        p = self.createProduct(template)
        self.registerProduct(index, p)
        return p

    @abstractmethod
    def createProduct(self, template: AbstructSdoDataBody.__class__):
        pass

    @abstractmethod
    def registerProduct(self, index: int, product):
        pass


@dataclass
class ConcreteSDODataFactory(SDODataFactory):
    """SDO Data Factory 具象クラス"""

    entries: Dict[int, AbstructSdoDataBody] = field(default_factory=dict, init=False)
    """オブジェクトディクショナリ
        
    Args:
        (int): SDOインデックス
        (SdoDataBody): SDOデータオブジェクト
    """

    def createProduct(self, template: AbstructSdoDataBody.__class__):
        """テンプレートクラスからインスタンスを作成

        Args:
            template(class): テンプレートクラス

        Return:
            SDODataBody: 生成したインスタンス
        """
        return template()

    def registerProduct(self, index: int, product: AbstructSdoDataBody):
        """インデックス毎のSDODataBodyオブジェクトが格納された辞書に登録"""
        self.entries[index] = product


@dataclass
class SdoEntry(Generic[T]):
    """SDOデータコンテナ

    ジェネリクス型変数を持つため、インスタンス定義時には型を指定すること。
    """

    name: str = field(default_factory=str, init=False)
    """SDO Entry名"""
    sub_index: int
    """SDO サブインデックス"""
    value: T
    """SDO 値"""
    format: str
    """
    struct.unpackで指定するformatの文字列
    
    https://docs.python.org/ja/3/library/struct.html#format-characters
    """
    size: int = field(default=None, init=True)
    """データサイズ（Byte）データ型により指定するべきサイズの意味が異なる。

    プリミティブ型:
        :mod:`format` で指定したサイズ。例えば5つの連続した32bit整数であればformatには ``5I`` と指定され、サイズは5 x 4 = 20byteが得られる。
    コレクション型:
        要素数 * :mod:`format` で指定したサイズ
    """
    enable: bool = field(default=False, init=True)
    """Enabled flag by SDO information service"""

    def __post_init__(self):
        if self.size is None:
            if is_primitive(self.value):
                self.size = calcsize(self.format)
            else:
                self.size = len(self.value) * calcsize(self.format)


@dataclass
class SdoDataBody(AbstructSdoDataBody):
    """SDODataBody実装クラス

    Raises:
        TypeError: SdoEntry以外のメンバを登録した場合に発生

    """

    def __post_init__(self):
        for each_field in self.__annotations__.items():
            if not issubclass(each_field[1], SdoEntry):
                raise TypeError(
                    f"Can't define as field except 'SdoDataBody'. Actual {each_field[0]} : {each_field[1]}"
                )
        # self.sub_index_dic = {asdict(self)[k]['sub_index']: k for k in asdict(self)}

    @property
    def unpack_format(self) -> str:
        """struct.unpack の formatを生成する"""
        # dataの要素のサイズがフォーマットの規定サイズより大きければ規定サイズ個数を頭に付け、そうでなければ
        result = "="
        for field in dataclasses.fields(self):
            entry = getattr(self, field.name)
            temp = ""
            format_size = calcsize(entry.format)
            if not entry.enable:
                continue
            if is_primitive(entry.value):
                # for primitive entry value
                logger.debug(f"{entry.value} : is primitive.")
                if format_size < entry.size:
                    temp += str(int(entry.size / format_size)) + entry.format
                else:
                    temp += entry.format
            elif hasattr(entry.value, "__iter__"):
                # for iterable object entry
                temp += str(int(entry.size / format_size)) + entry.format
                logger.debug(
                    f"{entry.value} : is NOT primitive. size :{entry.size}, format size: {format_size}, format: {temp}"
                )
            else:
                temp += entry.format
            if calcsize(result) % 2 == 0 or calcsize(result + temp) % 2 == 0:
                result += temp
            else:
                result += "x" + temp
        return result

    @property
    def total_size(self) -> int:
        return sum([getattr(self, f.name).size for f in dataclasses.fields(self) if getattr(self, f.name).enable])

    @property
    def values(self):
        dict_value = asdict(self)
        return {f: dict_value[f]["value"] for f in dict_value}

    def set_value(self, sub_index: int, value: SdoEntry):
        # sub_index_dic = {asdict(self)[k]['sub_index']: k for k in asdict(self)}
        # if sub_index in sub_index_dic:
        #    setattr(self, sub_index_dic[sub_index], value)
        # else:
        #    raise ValueError(f"No such subindex :{sub_index}")
        for item in dataclasses.fields(self):
            if getattr(self, item.name).sub_index == sub_index:
                setattr(self, item.name, value)
        raise ValueError(f"No such subindex :{sub_index}")

    def get_value(self, sub_index: int) -> SdoEntry:
        for item in dataclasses.fields(self):
            print(getattr(self, item.name).sub_index)
            if getattr(self, item.name).sub_index == sub_index:
                return getattr(self, item.name)

        raise ValueError(f"No such sub index :{sub_index}, attribute: {item.name}, class:{self.__class__.__name__}")


@dataclass
class SdoMetadata:
    """SDOメタデータ"""

    index: int
    """インデックス番号"""
    sub_index: int
    """サブインデックス番号"""
    support_complete_access: bool
    """complete access可能エントリ"""
    max_sub_index: int
    """最大サブインデックス番号"""
    request_container: Structure.__class__
    """アップロードコマンドのリクエストデータを格納するctype.Structureを基底クラスとしたコンテナクラスを定義"""
    response_container: SdoDataBody.__class__
    """アップロードコマンドのレスポンスデータを格納するSdoDataBodyを基底クラスとしたコンテナクラスを定義"""


@dataclass
class MappingMember:
    """ODマッピング定義オブジェクト"""

    index_range: Tuple[int, int]
    """ODのインデックス範囲"""
    metadata: SdoMetadata
    """ :obj:`SDOメタデータ <pyetg1510.mailbox.sdo_data_factory.SdoMetadata>` """


class SdoMetadataMapper:
    """SDOデータマッピングクラス"""

    def __post_init__(self):
        """メンバにMappingMember以外のデータが含まれていたらValueErrorとする"""
        for each_field in self.__annotations__.items():
            if not issubclass(each_field[1], MappingMember):
                raise TypeError(
                    f"Can't define as field except '{MappingMember.__class__.__name__}'. Actual {each_field[0]} : {each_field[1]}"
                )

    @classmethod
    def find(cls, index: int) -> MappingMember:
        """指定したインデックスに応じたODマッピング定義オブジェクトを返す

        Args:
            index(int): SDOインデックス番号

        Return:
            MappingMember: ODマッピング定義オブジェクト

        """
        for field in vars(cls):
            if isinstance(vars(cls)[field], MappingMember):
                index_range = vars(cls)[field].index_range
                if index_range[0] <= index <= index_range[1]:
                    return vars(cls)[field]
        return None

    @classmethod
    def find_start(cls, index: int):
        return cls.find(index).index_range[0]


class ODList(SdoDataBody):
    """ODリストのメンバー"""
    ListType: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H", enable=True))
    ObjectIndex: SdoEntry = field(
        default_factory=lambda: SdoEntry[list[int]](sub_index=0, value=[0], format="H", enable=True)
    )


ODListFormat = SdoMetadata(
    index=-1,
    sub_index=-1,
    support_complete_access=False,
    max_sub_index=0,
    request_container=SDOInformationODListRequest,
    response_container=ODList,
)


class SDOInfoDescription(SdoDataBody):
    """個々のSDO Indexの詳細情報"""
    Index: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H", enable=True))
    DataType: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H", enable=True))
    MaxSubindex: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B", enable=True))
    ObjectCode: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B", enable=True))
    Name: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=0, value="", format="s", enable=True))


SDOInfoDescriptionFormat = SdoMetadata(
    index=0,
    sub_index=-1,
    support_complete_access=False,
    max_sub_index=0,
    request_container=SDOInformationDescriptionRequest,
    response_container=SDOInfoDescription,
)


class SDOInfoEntry(SdoDataBody):
    """SDOエントリ情報"""
    Index: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H", enable=True))
    Subindex: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B", enable=True))
    ValueInfo: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="B", enable=True))
    DataType: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H", enable=True))
    BitLength: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H", enable=True))
    ObjectAccess: SdoEntry = field(
        default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="H", enable=True)
    )
    Data: SdoEntry = field(default_factory=lambda: SdoEntry[str](sub_index=0, value="", format="s", enable=True))


SDOInfoEntryFormat = SdoMetadata(
    index=0,
    sub_index=0,
    support_complete_access=False,
    max_sub_index=0,
    request_container=SDOInformationEntryRequest,
    response_container=SDOInfoEntry,
)


class SDOInfoError(SdoDataBody):
    AbortCode: SdoEntry = field(default_factory=lambda: SdoEntry[int](sub_index=0, value=0, format="I", enable=True))


SDOInfoErrorFormat = SdoMetadata(
    index=0,
    sub_index=0,
    support_complete_access=False,
    max_sub_index=0,
    request_container=None,
    response_container=SDOInfoError,
)


@dataclass
class SdoDataController:
    """SDO メッセージサービス

    Args:
        session(EtherCATMasterConnection): 通信コネクタオブジェクト
        get_info(bool): SDO Information serviceの問い合わせ時はTrueにする
    """

    session: EtherCATMasterConnection
    sdo_data: SdoDataBody = field(default=None)
    get_info: bool = field(default=False)

    def __post_init__(self):
        self.data_body_size: int = 0
        self.index_counter: int = 0

    def _map(self, raw_data: bytes):
        """SDOデータ本体部分の内部モデルへのマッピング関数

        Unpack received data and put there on each member of 'values' array at the 'SdoFormat' instance as a 'SdoEntry' data type.
        受信したバイトデータをアンパックし、 :meth:`get_object_dictionary <pyetg1510.etg_1510.MasterODSpecification.get_object_dictionary>` で作成したSdoFormat

        Args:
            raw_data(bytes): Upload commandで受信したレスポンスデータのデータ本体部分
            self.sdo_data(SdoDataBody): マッピング先のデータコンテナ
        Raises:
            TypeError: struct.unpackにおけるformat指定が不正な場合、または、サイズが合わない場合。
            struct.error: struct.unpack 処理が失敗した場合。
        """
        logger.info(self.sdo_data)
        logger.info(
            f"Before adjusting size, Unpack format:{self.sdo_data.unpack_format}, format size:{calcsize(self.sdo_data.unpack_format)}, Setting size: {self.sdo_data.total_size}, Actual size: {len(raw_data)}"
        )
        if self.sdo_data.total_size < len(raw_data):
            lastkey = next(reversed(asdict(self.sdo_data)), None)
            current_value = getattr(self.sdo_data, lastkey)
            current_value.size += len(raw_data) - self.sdo_data.total_size
            setattr(self.sdo_data, lastkey, current_value)
        unpack_format = self.sdo_data.unpack_format
        format_size = calcsize(unpack_format)
        if format_size <= 0:
            raise ValueError(f"All member are disabled. ({unpack_format}) Nothing to fetch data.")

        logger.info(
            f"Unpack format: {unpack_format}, Unpack size: {format_size}, Calculated size: {self.sdo_data.total_size},raw bytes data size : {len(raw_data)}"
        )
        if len(raw_data) < format_size:
            logger.error(f"Data: {raw_data}, Unpack format: {unpack_format}")
            raw_data += b"\0" * (format_size - len(raw_data))
            # raise ValueError(f"Required data size at least {format_size} byte. actual: {len(raw_data)} byte")
        try:
            _data = list(unpack(unpack_format, raw_data))
        except error as e:
            logger.exception(
                f"""unpack error: unpack format size: {calcsize(unpack_format)}, data size: {len(raw_data)}
    Specified datamodel: {self.sdo_data}
            """
            )

            raise
        _data = [f.strip(b"\0").decode() if type(f) is bytes else f for f in _data]
        sdo_data_dict = asdict(self.sdo_data)
        sdo_data_dict = {k: sdo_data_dict[k] for k in sdo_data_dict if sdo_data_dict[k]["enable"]}
        if len(sdo_data_dict) > len(_data):
            raise ValueError(
                f"""Mismatching configured SDO data number and received data.
    Received data: count: {len(_data)}, unpack format {unpack_format}, Unpack data: {_data}
    Configured model: count: {len(sdo_data_dict)}, model:{sdo_data_dict}"""
            )
        k = 0
        i = 0
        for key in sdo_data_dict:
            if is_primitive(getattr(self.sdo_data, key).value):
                value = getattr(self.sdo_data, key)
                value.value = _data[i]
                setattr(self.sdo_data, key, value)
                i += 1
            elif type(getattr(self.sdo_data, key).value) is type(_data):
                byte_size = getattr(self.sdo_data, key).size
                list_size = int(byte_size / calcsize(getattr(self.sdo_data, key).format))
                set_list = _data[i : i + list_size]
                value = getattr(self.sdo_data, key)
                value.value = set_list
                setattr(self.sdo_data, key, value)
                i += len(set_list)
            else:
                raise TypeError(
                    f"Type unmatched. \n {self.sdo_data[key].name} :type(self.sdo_data[key].value)\n specified:{type(_data[i])}"
                )

            k += 1

    def _object_initialization(self, sdo_metadata: SdoMetadata):
        """リクエスト、レスポンス各メッセージコンテナを初期化する。"""

        if self.get_info:
            self.response_message = SDOResponseMessage(sdo_service=SdoService.INFO)
            self.request_message = SDORequestInfoMessage(sdo_service=SdoService.INFO)
            if sdo_metadata == ODListFormat:
                logger.info("Fetching OD List")
                self.request_message.opcode = SdoInfoOpcode.GET_OD_LIST_REQ
            elif sdo_metadata == SDOInfoDescriptionFormat:
                logger.info("Fetching Object Description")
                self.request_message.opcode = SdoInfoOpcode.GET_DESCRIPTION_REQ
            elif sdo_metadata == SDOInfoEntryFormat:
                logger.info("Fetching Entry Description")
                self.request_message.opcode = SdoInfoOpcode.GET_ENTRY_REQ
        else:
            self.response_message = SDOResponseMessage(sdo_service=SdoService.RESPONSE)
            self.request_message = SDOCommandMessage(sdo_service=SdoService.REQUEST)

        self.request_message.sdo_command_data = sdo_metadata.request_container()

    async def fetch(self, sdo_metadata: SdoMetadata, sdo_data: SdoDataBody):
        """SDO Upload リクエストを発行し、SdoDataBodyモデルへマッピングする

        Args:
            sdo_metadata(SdoMetaData): :obj:`SDOメタデータ <pyetg1510.mailbox.sdo_data_factory.SdoMetadata>`
            sdo_data(SdoDataBody): 受信したSDOデータを格納するコンテナオブジェクト
        """
        self.sdo_data = sdo_data
        self._object_initialization(sdo_metadata=sdo_metadata)
        self.request_message.index = sdo_metadata.index
        self.request_message.sub_index = sdo_metadata.sub_index
        self.request_message.complete_access = sdo_metadata.support_complete_access

        logger.debug(
            f"Command to be requested index:{self.request_message.index}, subindex: {self.request_message.sub_index}, complete access: {self.request_message.complete_access}"
        )
        # request and wait response

        await self.session.send_data(self.request_message.make_request_frame())
        # parse until CoE header message
        self.response_message.parse_response_frame(self.session.received_data)

        # Parse SDO message
        # 1. Make sure data size either specified size or default size by SizeIndicator
        # 2. If SizeIndicator is True and expedited bit is True, data size is specified at 2bit.
        data_body_offset = 0
        if (
            "SizeIndicator" in [f[0] for f in self.response_message.sdo_header._fields_]
            and self.response_message.sdo_header.SizeIndicator == 1
        ):
            if self.response_message.sdo_header.TransferType == 1:
                # case : SDO upload expedited response]
                self.data_body_size = 4 - self.response_message.sdo_header.DataSetSize
                logger.debug(
                    f"Expedited : True, size specified: {self.response_message.sdo_header.DataSetSize}, Calculated size:{self.data_body_size}"
                )
            else:
                temp = unpack_from("@I", self.response_message.data_body, 0)
                self.data_body_size = temp[0]
                data_body_offset = 4
        else:
            self.data_body_size = 4
        if len(self.response_message.data_body[data_body_offset:]) < self.data_body_size:
            logger.warning("!!!STOP!!! Size too low")
            self.index_counter = 0
            raise StopAsyncIteration()
        # SDO Information service checkking Opcode
        if self.get_info and (
            "Opcode" in [f[0] for f in self.response_message.sdo_header._fields_]
            and self.response_message.sdo_header.Opcode == SdoInfoOpcode.SDO_INFO_ERR_REQ.value
        ):
            self.sdo_data = SDOInfoErrorFormat.response_container()
            sdo_metadata = SDOInfoErrorFormat

        # Mapping SDO data body to native model
        # try:
        logger.debug(f"SDO Body message {self.response_message.data_body[data_body_offset:]}")
        self._map(raw_data=self.response_message.data_body[data_body_offset:])
        logger.debug(f"mapped data: {self.sdo_data}")
        # except (ValueError, TypeError, TimeoutError, asyncio.exceptions.CancelledError, asyncio.exceptions.InvalidStateError) as e:
        #    logger.warning(e)
        #    self.index_counter = 0
        #    raise StopAsyncIteration()
        logger.debug(
            f"Sequence number: {hex(sdo_metadata.index)}:{hex(sdo_metadata.sub_index)} / {self.index_counter}"
        )
        self.index_counter += 1
        # return copy.deepcopy(self.sdo_data)



from .helper.logging_service import *
from .helper.settings import *
from .etg_1510 import *
from .sdo_1xxx_master_object import *
from .sdo_8xxx_configuration_data import *
from .sdo_9xxx_information_data import *
from .sdo_axxx_master_diagnosis import *
from .sdo_fxxx_controls import *

VERSION = (0, 0, 1)

__version__ = ".".join([str(x) for x in VERSION])



"""Mailbox Gateway 通信フレーム制御モジュール

"""
import sys
from ctypes import *
from bitarray import bitarray, util
from dataclasses import dataclass, field
from enum import Enum
from typing import Union
from pyetg1510.helper import SysLog

logger = SysLog.logger


class EtherCATProtocolType(Enum):
    PDU = 0x01
    """Protocol Data Unit"""
    NETWORK_VARIABLES = 0x04
    MAILBOX = 0x05


class SdoService(Enum):
    REQUEST = 0x02
    RESPONSE = 0x03
    INFO = 0x08


class SdoRequestCommand(Enum):
    SDOREQ_DOWNLOAD_SEGMENTED = 0x00
    """Req/Download Segmented Res/Upload Segmented"""
    SDOREQ_DOWNLOAD = 0x01
    """Req/Download Res/Download Segmented Info/Get OD List Req."""
    SDOREQ_UPLOAD = 0x02
    """Req/Upload Res/Upload Info/Get OD List Resp."""
    SDOREQ_UPLOAD_SEGMENTED = 0x03
    """Req/Upload Segmented Res/Download Info/Get Object Description Req."""
    SDOREQ_ABORT_TRANSFER = 0x04
    """Req/Abort Transfer Info./Get Object Description Resp."""


class SdoInfoOpcode(Enum):
    GET_OD_LIST_REQ = 0x01
    GET_OD_LIST_RES = 0x02
    GET_DESCRIPTION_REQ = 0x03
    GET_DESCRIPTION_RES = 0x04
    GET_ENTRY_REQ = 0x05
    GET_ENTRY_RES = 0x06
    SDO_INFO_ERR_REQ = 0x07


class SdoResponseCommand(Enum):
    SDORES_UPLOAD_SEGMENTED = 0x00
    """Req/Download Segmented Res/Upload Segmented"""
    SDORES_DOWNLOAD_SEGMENTED = 0x01
    """Req/Download Res/Download Segmented Info/Get OD List Req."""
    SDORES_UPLOAD = 0x02
    """Req/Upload Res/Upload Info/Get OD List Resp."""
    SDORES_DOWNLOAD = 0x03
    """Req/Upload Segmented Res/Download Info/Get Object Description Req."""


class SdoInfoCommand(Enum):
    SDOINFO_ODLIST_REQ = 0x01
    """Req/Download Res/Download Segmented Info/Get OD List Req."""
    SDOINFO_ODLIST_RES = 0x02
    """Req/Upload Res/Upload Info/Get OD List Resp."""
    SDOINFO_DESCRIPTION_REQ = 0x03
    """Req/Upload Segmented Res/Download Info/Get Object Description Req."""
    SDOINFO_DESCRIPTION_RES = 0x04
    """Req/Abort Transfer Info./Get Object Description Resp."""
    SDOINFO_ENTRY_REQ = 0x05
    """Info/Get Entry Description."""
    SDOINFO_ENTRY_RES = 0x06
    """Info/Get Entry Description Resp."""
    SDOINFO_INFO_ERR_REQ = 0x07
    """Info/SDO Info Error Req."""


class MailBoxFrameOffsetAddress(Enum):
    ETHERCAT_HEADER = 0
    MAILBOX_HEADER = 2
    COE_HEADER = 8
    SDO_HEADER = 10
    SDO_DATA = 14


class EtherCATHeader(Structure):
    _pack_ = 1
    _fields_ = [
        ("Length", c_uint16, 11),  # 'H' -> int
        ("Reserved", c_uint8, 1),  # '?' -> bool
        ("DataType", c_uint8, 4),  # 'B' -> int
    ]


class MailboxHeader(Structure):
    _pack_ = 1
    _fields_ = [
        ("Length", c_uint16, 16),  # 'H' -> int
        ("Address", c_uint16, 16),  # 'H' -> int
        ("Channel", c_uint8, 6),  # 'B' -> int
        ("Prio", c_uint8, 2),  # 'B' -> int
        ("Type", c_uint8, 4),  # 'B' -> int
        ("Cnt", c_uint8, 3),  # 'B' -> int
        ("Reserved", c_uint8, 1),  # '?' -> bool
    ]


class CoEHeader(Structure):
    _pack_ = 1
    _fields_ = [
        ("Number", c_uint16, 9),  # 'H' -> int
        ("Reserved", c_uint8, 3),  # 'B' -> int
        ("Service", c_uint8, 4),  # 'B' -> int
    ]


class SDORequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("Reserved", c_uint8, 4),  # 'H' -> int
        ("CompleteAccess", c_uint8, 1),  # '?' -> bool
        ("CommandSpecifier", c_uint8, 3),  # 'B' -> int
        ("Index", c_uint16, 16),  # 'H' -> int
        ("SubIndex", c_uint8, 8),  # 'B' -> int
        ("Reserved2", c_uint32, 32),  # 'H' -> int
    ]


class SDOResponse(Structure):
    _pack_ = 1
    _fields_ = [
        ("SizeIndicator", c_uint8, 1),  # True : enable following size specification
        ("TransferType", c_uint8, 1),  # True: Expedited / False: Normal
        ("DataSetSize", c_uint8, 2),  # 0: 4byte, 1 : 3byte, 2: 2byte, 3: 1byte
        ("CompleteAccess", c_uint8, 1),  # multi subindex
        ("CommandSpecifier", c_uint8, 3),  # defined as 'SdoResponseCommand' enum type
        ("Index", c_uint16, 16),  # 'H' -> int
        ("SubIndex", c_uint8, 8),  # 'B' -> int
    ]


class SDOInformationHeader(Structure):
    _pack_ = 1
    _fields_ = [
        ("Opcode", c_uint8, 7),  #
        ("Incomplete", c_uint8, 1),  #
        ("Reserved", c_uint8, 8),  #
        ("FragmentsLeft", c_uint16, 16),  #
    ]


class SDOInformationODListRequest(Structure):
    _fields_ = [("ListType", c_uint16, 16)]


class SDOInformationDescriptionRequest(Structure):
    _fields_ = [
        ("Index", c_uint16, 16),
    ]


class SDOInformationEntryRequest(Structure):
    _fields_ = [("Index", c_uint16, 16), ("Subindex", c_uint8, 8), ("ValueInfo", c_uint8, 8)]


@dataclass
class SDOMessage:
    sdo_service: SdoService = field(default_factory=SdoService, init=True)
    station_address: int = field(default=0x0000, init=True)  # Default: Master
    mailbox_type: int = field(default=0x03, init=True)  # Default: CoE
    session_counter: int = field(default=1, init=False)
    index: int = field(default_factory=int, init=False)
    sub_index: int = field(default_factory=int, init=False)
    complete_access: bool = field(default_factory=bool, init=False)
    sdo_command: Union[SdoRequestCommand, SdoResponseCommand] = field(
        default=SdoRequestCommand.SDOREQ_UPLOAD, init=False
    )
    ethercat_header: EtherCATHeader = field(default_factory=EtherCATHeader, init=False)
    mailbox_header: MailboxHeader = field(default_factory=MailboxHeader, init=False)
    coe_header: CoEHeader = field(default_factory=CoEHeader, init=False)

    def __new__(cls, *args, **kwargs):
        dataclass(cls)
        return super().__new__(cls)

    def __post_init__(self):
        if self.sdo_service == SdoService.REQUEST:
            self.sdo_header = SDORequest()
        elif self.sdo_service == SdoService.RESPONSE:
            self.sdo_header = SDOResponse()
        elif self.sdo_service == SdoService.INFO:
            self.sdo_header = SDOInformationHeader()

    def get_bytes(self, source: Structure) -> bitarray:
        """Get bitarray object"""

        # Works for either Python2 or Python3

        _bitarray = bitarray(endian=sys.byteorder)
        _bitarray[:] = 0

        for item in source._fields_:
            a = bitarray(endian=sys.byteorder)
            a[:] = 0
            value = getattr(source, item[0])
            a.frombytes(value.to_bytes(length=1 + int(item[2] / 8), byteorder=sys.byteorder))

            if a.endian() == "big":
                a = a[0 - item[2] :]
                _bitarray = a + _bitarray
            else:
                a = a[0 : item[2]]
                _bitarray = _bitarray + a

        if _bitarray.endian() != "big":
            _bitarray.bytereverse()

        return _bitarray

    def make_coe_header(self, sdo_data_size: int, increase_session: bool = True) -> bitarray:
        # CoE Header message parameter set
        self.coe_header.Number = 0
        self.coe_header.Reserved = 0
        self.coe_header.Service = self.sdo_service.value
        coe_header_bytes = self.get_bytes(self.coe_header)
        body_size = int((sdo_data_size + len(coe_header_bytes)) / 8)
        self.mailbox_header.Length = body_size
        self.mailbox_header.Address = self.station_address
        self.mailbox_header.Channel = 0
        self.mailbox_header.Prio = 0
        self.mailbox_header.Type = self.mailbox_type

        # Mailbox Header message parameter set
        if self.session_counter > 7 or self.session_counter <= 0:
            raise ValueError(f"session_counter must be within 1-7. Set to {self.session_counter}")

        if increase_session:
            if self.session_counter >= 7 or self.session_counter <= 0:
                self.session_counter = 1
            else:
                self.session_counter += 1

        self.mailbox_header.Cnt = self.session_counter
        self.mailbox_header.Reserved = 0
        mailbox_header_bytes = self.get_bytes(self.mailbox_header)

        # EtherCAT Header message parameter set
        total_length = int(len(mailbox_header_bytes) / 8) + body_size

        self.ethercat_header.Length = total_length
        self.ethercat_header.Reserved = 0
        self.ethercat_header.DataType = EtherCATProtocolType.MAILBOX.value
        ecat_header_bytes = self.get_bytes(self.ethercat_header)

        # join whole frames
        _bitarray = bitarray()
        _bitarray.extend(ecat_header_bytes)
        _bitarray.extend(mailbox_header_bytes)
        _bitarray.extend(coe_header_bytes)

        return _bitarray


class SDOResponseMessage(SDOMessage):
    """deserialize received byte message until coe header.
    If you find deserialize logic for sdo message body, you should see map method at the PySdoDataMode.SdoController
    """

    def __post_init__(self):
        super().__post_init__()
        if self.sdo_service not in [SdoService.RESPONSE, SdoService.INFO]:
            raise ValueError("Property sdo_service should be only in [Sdoservice.RESPONSE, SdoService.INFO]")

    def parse_response_frame(self, response: bytes):
        def debug_log(s: Structure):
            result = {f[0]: getattr(s, f[0]) for f in s._fields_}
            logger.debug(f"{self.__class__.__name__} {s.__class__.__name__}: {result}")

        _bitarray_response = bitarray()
        _bitarray_response.frombytes(response)

        # Parse EtherCAT header
        ethercat_header = _bitarray_response[
            MailBoxFrameOffsetAddress.ETHERCAT_HEADER.value * 8 : MailBoxFrameOffsetAddress.MAILBOX_HEADER.value * 8
        ]
        self._map_structure(ethercat_header, self.ethercat_header)

        debug_log(self.ethercat_header)

        # Parse Mailbox header
        mailbox_header = _bitarray_response[
            MailBoxFrameOffsetAddress.MAILBOX_HEADER.value * 8 : MailBoxFrameOffsetAddress.COE_HEADER.value * 8
        ]
        self._map_structure(mailbox_header, self.mailbox_header)

        debug_log(self.mailbox_header)

        # Parse CoE header
        coe_header = _bitarray_response[
            MailBoxFrameOffsetAddress.COE_HEADER.value * 8 : MailBoxFrameOffsetAddress.SDO_HEADER.value * 8
        ]
        self._map_structure(coe_header, self.coe_header)

        debug_log(self.coe_header)

        # Parse sdo response data
        if self.sdo_service == SdoService.INFO:
            sdo_header = _bitarray_response[
                MailBoxFrameOffsetAddress.SDO_HEADER.value * 8 : MailBoxFrameOffsetAddress.SDO_DATA.value * 8
            ]
            self._map_structure(sdo_header, self.sdo_header)

            data_body_bitarray = _bitarray_response[MailBoxFrameOffsetAddress.SDO_DATA.value * 8 :]

            self.data_body = data_body_bitarray.tobytes()

        else:
            sdo_header = _bitarray_response[
                MailBoxFrameOffsetAddress.SDO_HEADER.value * 8 : MailBoxFrameOffsetAddress.SDO_DATA.value * 8
            ]
            self._map_structure(sdo_header, self.sdo_header)

            data_body_bitarray = _bitarray_response[MailBoxFrameOffsetAddress.SDO_DATA.value * 8 :]

            self.data_body = data_body_bitarray.tobytes()

        debug_log(self.sdo_header)

    def _map_structure(self, input: bitarray, struct_obj: Structure):
        offset = 0
        if sys.byteorder != input.endian():
            input.bytereverse()
        for item in struct_obj._fields_:
            next_offset = offset + item[2]
            sliced = input[offset:next_offset]
            if sys.byteorder != input.endian():
                sliced.reverse()
            if len(sliced) > 0:
                setattr(struct_obj, item[0], util.ba2int(sliced))
            offset += item[2]


class SDOCommandMessage(SDOMessage):
    sdo_command_data: Structure = field(default=SDOInformationODListRequest(), init=False)

    def __post_init__(self):
        super().__post_init__()

    def make_request_frame(self, increase_session: bool = True) -> bytes:
        # SDO Request message parameter set
        if isinstance(self.sdo_header, SDORequest):
            self.sdo_header.Reserved = 0
            self.sdo_header.CompleteAccess = int(self.complete_access)
            self.sdo_header.CommandSpecifier = self.sdo_command.value
            self.sdo_header.Index = self.index
            self.sdo_header.SubIndex = self.sub_index
            self.sdo_header.Reserved2 = 0
        else:
            TypeError(f"sdo_header should be SDORequest type.")

        sdo_bitarray = self.get_bytes(self.sdo_header)
        if self.sdo_command_data is not None:
            sdo_bitarray.extend(self.get_bytes(self.sdo_command_data))

        logger.debug(
            f"Request Body: {self.sdo_command_data.__class__.__name__}, Index:{hex(self.index)}, Subindex:{hex(self.sub_index)} specified."
        )

        _bitarray = self.make_coe_header(len(sdo_bitarray))
        _bitarray.extend(sdo_bitarray)

        # convert bitarray with bytes data and return
        return _bitarray.tobytes()


class SDORequestInfoMessage(SDOMessage):
    """Build SDO "Info Service Data" frame in SDO information Service Request command"""
    opcode: SdoInfoOpcode = field(default=SdoInfoOpcode.GET_OD_LIST_REQ, init=True)
    sdo_command_data: Structure = field(default=None, init=False)

    def __post_init__(self):
        super().__post_init__()

    def make_request_frame(self, increase_session: bool = True) -> bytes:
        # SDO Request message parameter set
        if isinstance(self.sdo_header, SDOInformationHeader):
            self.sdo_header.Opcode = self.opcode.value
            self.sdo_header.Incomplete = 0
            self.sdo_header.FragmentsLeft = 0
            self.sdo_header.ListType = 0x01
            if isinstance(self.sdo_command_data, SDOInformationODListRequest):
                self.sdo_command_data.ListType = 0x01
            elif isinstance(self.sdo_command_data, SDOInformationDescriptionRequest):
                self.sdo_command_data.Index = self.index
            elif isinstance(self.sdo_command_data, SDOInformationEntryRequest):
                self.sdo_command_data.Index = self.index
                self.sdo_command_data.Subindex = self.sub_index
                self.sdo_command_data.ValueInfo = 0x7F

        else:
            TypeError(f"sdo_header should be SDOInformationHeader type.")

        logger.info(
            f"Request Body: {self.sdo_command_data.__class__.__name__}, Index:{hex(self.index)}, Subindex:{hex(self.sub_index)} specified."
        )

        sdo_bitarray = self.get_bytes(self.sdo_header)
        if self.sdo_command_data is not None:
            sdo_bitarray.extend(self.get_bytes(self.sdo_command_data))

        _bitarray = self.make_coe_header(len(sdo_bitarray))
        _bitarray.extend(sdo_bitarray)

        # convert bitarray with bytes data and return
        return _bitarray.tobytes()
